@@ -2,7 +2,7 @@
 
 # MIT License
 # 
-# Copyright (c) 2020 Tencent Cloud.
+# Copyright (c) 2021 Handle.
 # 
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -31,7 +31,10 @@ Example:
     print(await functions.invoke_async('hello'))
 '''
 
+import sys
 import time
+import json
+import base64
 import asyncio
 import inspect
 import threading
@@ -54,6 +57,8 @@ from tencent.cloud.serverless.functions.client import FunctionType
 from tencent.cloud.serverless.functions.client import FunctionRuntime
 from tencent.cloud.serverless.functions.client import FunctionResultType
 
+from tencent.cloud.serverless.functions.integrate import IntegrateDispatch
+
 class FunctionResultFuture(asyncio.Future):
     '''
     A Future type that represents the result of a
@@ -398,6 +403,92 @@ class Client(client.AbstractClient):
             return (self._get_event_loop().run_until_complete(self.invoke_async(region_id,
                 namespace_name, function_name, function_event, function_version, True))['request_id'])
 
+    async def routine_invoke_async(self,
+        region_id: str,
+        namespace_name: str,
+        function_name: str,
+        routine_name: str,
+        routine_parameter: dict = None,
+        function_version: str = None,
+        function_async: bool = False
+    ) -> object:
+        '''
+        Invoke given the specified routine of the cloud function.
+
+        Args:
+            region_id: Data center unique identifier.
+            namespace_name: Namespace name.
+            function_name: Serverless cloud function name.
+            routine_name: Routine name for serverless cloud functions.
+            routine_parameter: Routine parameters for serverless cloud functions.
+            function_version: Serverless cloud function version name.
+            function_async: Make Serverless cloud function asynchronous invoke.
+        
+        Returns:
+            Returns the actual return value of the given routine for the
+                given serverless cloud function.
+
+        Raises:
+            ValueError: Parameter values are not as expected.
+            InvokeError: Invoke Cloud Function failed.
+            ActionError: Invoke Cloud Function error.
+        '''
+
+        if not routine_name or not isinstance(routine_name, str):
+            raise ValueError('<routine_name> value invalid')
+
+        if routine_parameter and not isinstance(routine_parameter, dict):
+            raise ValueError('<routine_parameter> value invalid')
+
+        return await self.easy_invoke_async(region_id, namespace_name,
+            function_name, {
+                'protocol': {
+                    'version': 1,
+                    'payload': base64.b64encode(json.dumps(
+                        {
+                            'routine_name': routine_name,
+                            'routine_parameter': routine_parameter
+                        }
+                    ).encode()).decode()
+                }
+            }, function_version, function_async)
+    
+    def routine_invoke(self,
+        region_id: str,
+        namespace_name: str,
+        function_name: str,
+        routine_name: str,
+        routine_parameter: dict = None,
+        function_version: str = None,
+        function_async: bool = False
+    ) -> object:
+        '''
+        Invoke given the specified routine of the cloud function.
+
+        Args:
+            region_id: Data center unique identifier.
+            namespace_name: Namespace name.
+            function_name: Serverless cloud function name.
+            routine_name: Routine name for serverless cloud functions.
+            routine_parameter: Routine parameters for serverless cloud functions.
+            function_version: Serverless cloud function version name.
+            function_async: Make Serverless cloud function asynchronous invoke.
+        
+        Returns:
+            Returns the actual return value of the given routine for the
+                given serverless cloud function.
+
+        Raises:
+            ValueError: Parameter values are not as expected.
+            InvokeError: Invoke Cloud Function failed.
+            ActionError: Invoke Cloud Function error.
+        '''
+
+        return self._get_event_loop().run_until_complete(self.routine_invoke_async(
+            region_id, namespace_name, function_name, routine_name,
+            routine_parameter, function_version, function_async
+        ))
+
     async def select_function_async(self,
         region_id: str,
         namespace_name: str,
@@ -448,6 +539,64 @@ class Client(client.AbstractClient):
         return (lambda **function_event: self.easy_invoke(region_id,
             namespace_name, function_name, function_event, function_version, function_async))
 
+    async def select_routine_async(self,
+        region_id: str,
+        namespace_name: str,
+        function_name: str,
+        routine_name: str,
+        function_version: str = None,
+        function_async: bool = False
+    ):
+        '''
+        Creates a Python native asynchronous function for a
+            given routine given a serverless cloud function.
+
+        Args:
+            region_id: Data center unique identifier.
+            namespace_name: Namespace name.
+            function_name: Serverless cloud function name.
+            routine_name: Routine name for serverless cloud functions.
+            function_version: Serverless cloud function version name.
+            function_async: Make Serverless cloud function asynchronous invoke.
+
+        Returns:
+            Returns a Python native asynchronous function instance.
+        '''
+
+        return (lambda **routine_parameter: self.routine_invoke_async(
+            region_id, namespace_name, function_name, routine_name,
+            routine_parameter, function_version, function_async
+        ))
+    
+    def select_routine(self,
+        region_id: str,
+        namespace_name: str,
+        function_name: str,
+        routine_name: str,
+        function_version: str = None,
+        function_async: bool = False
+    ):
+        '''
+        Creates a Python native synchronous function for a
+            given routine given a serverless cloud function.
+
+        Args:
+            region_id: Data center unique identifier.
+            namespace_name: Namespace name.
+            function_name: Serverless cloud function name.
+            routine_name: Routine name for serverless cloud functions.
+            function_version: Serverless cloud function version name.
+            function_async: Make Serverless cloud function asynchronous invoke.
+
+        Returns:
+            Returns a Python native synchronous function instance.
+        '''
+
+        return (lambda **routine_parameter: self.routine_invoke(
+            region_id, namespace_name, function_name, routine_name,
+            routine_parameter, function_version, function_async
+        ))
+
     def bind_function(self,
         region_id: str,
         namespace_name: str,
@@ -479,12 +628,15 @@ class Client(client.AbstractClient):
         
         Returns:
             Returns the Python decorator handler function.
+        
+        Raises:
+            TypeError: The bound object is not supported.
         '''
 
         def decorator_handler(
             bound_function: object
         ):
-            if not callable(bound_function):
+            if not bound_function or not callable(bound_function):
                 raise TypeError('invalid binding object type')
 
             def invoke_handler(*args, **kwargs):
@@ -514,6 +666,64 @@ class Client(client.AbstractClient):
 
         return decorator_handler
 
+    def bind_routine(self,
+        region_id: str,
+        namespace_name: str,
+        function_name: str,
+        function_version: str = None,
+        function_async: bool = False,
+        routine_name: str = None
+    ) -> object:
+        '''
+        Binds a given Python native synchronous or asynchronous function
+            to a given routine for a given serverless cloud function.
+        
+        Args:
+            region_id: Data center unique identifier.
+            namespace_name: Namespace name of serverless cloud function.
+            function_name: Serverless cloud function name.
+            function_version: Serverless cloud function version name.
+            function_async: Make Serverless Cloud Function asynchronous invoke.
+            routine_name: Routine name for serverless cloud functions.
+        
+        Raises:
+            ValueError: Parameter values are not as expected.
+            TypeError: The bound object is not supported.
+        '''
+
+        if routine_name and not isinstance(routine_name, str):
+            raise ValueError('<routine_name> value invalid')
+
+        def decorator_handler(
+            bound_function: object
+        ) -> object:
+            if not bound_function or not callable(bound_function):
+                raise TypeError('invalid binding object type')
+
+            if not inspect.isfunction(bound_function):
+                raise ValueError('invalid binding object type')
+
+            method_instance: object = (self.routine_invoke_async if
+                inspect.iscoroutinefunction(bound_function) else self.routine_invoke)
+
+            def invoke_handler(*args, **kwargs):
+                routine_parameter: dict = dict()
+                parameter_names: list = inspect.getfullargspec(bound_function).args
+
+                for index, value in enumerate(args):
+                    routine_parameter[parameter_names[index]] = value
+
+                for name in kwargs:
+                    routine_parameter[name] = kwargs[name]
+
+                return method_instance(region_id, namespace_name, function_name,
+                    routine_name if routine_name else bound_function.__name__,
+                    routine_parameter, function_version, function_async)
+
+            return invoke_handler
+
+        return decorator_handler
+
     def _schedule_created_callback(self,
         function_schedule: FunctionSchedule
     ):
@@ -714,6 +924,12 @@ def invoke(
         ActionError: Invoke Cloud Function error.
     '''
     
+    if not region_id:
+        region_id = helper.get_cloud_function_region_id()
+    
+    if not namespace_name:
+        namespace_name = helper.get_cloud_function_namespace_name()
+
     return fetch_client().easy_invoke(region_id, namespace_name,
         function_name, function_event, function_version, function_async)
 
@@ -751,7 +967,52 @@ async def invoke_async(
         InvokeError: Invoke Cloud Function failed.
         ActionError: Invoke Cloud Function error.
     '''
+
+    if not region_id:
+        region_id = helper.get_cloud_function_region_id()
     
+    if not namespace_name:
+        namespace_name = helper.get_cloud_function_namespace_name()
+
     return await fetch_client().easy_invoke_async(region_id,
         namespace_name, function_name, function_event,
         function_version, function_async)
+
+def use_routine_dispatcher(
+    override_handler: bool = True
+) -> IntegrateDispatch:
+    '''
+    Use the routine dispatch framework for integrated
+        invoke for current serverless cloud functions.
+
+    Note that this function will immediately add or override
+        the main function at the source of the call, unless
+        the parameter override_handler is set to False.
+
+    Args:
+        override_handler: Whether to create or override the
+            serverless cloud function main processing function
+            at the source of this function call.
+
+    Returns:
+        Returns an instance of the integrated invoke dispatcher.
+
+    Raises:
+        ValueError: Parameter values are not as expected.
+        ModuleNotFoundError: Can't fault the module object from which
+            the current function call originated.
+    '''
+
+    if override_handler == None or not isinstance(override_handler, bool):
+        raise ValueError('<override_handler> value invalid')
+
+    integrate_dispatch: IntegrateDispatch = IntegrateDispatch()
+
+    if override_handler:
+        try:
+            module_name: str = inspect.currentframe().f_back.f_globals['__name__']
+            setattr(sys.modules[module_name], 'main', integrate_dispatch.handler)
+        except KeyError:
+            raise ModuleNotFoundError('no call source module found')
+
+    return integrate_dispatch
